import unittest
from unittest.mock import patch

from requests.exceptions import ConnectionError as RequestsConnectionError

from test.util import mock_connection_func, check_error_message, check_startswith_error_message
from weaviate.backup.backup import Backup, STORAGE_NAMES
from weaviate.exceptions import (
    UnexpectedStatusCodeException,
    BackupFailedException,
)


class TestBackup(unittest.TestCase):
    @patch("weaviate.backup.backup.Backup.get_create_status")
    def test_create(self, mock_status):
        """
        Test the `create` method.
        """

        # error messages
        backup_type_err_msg = lambda dt: f"'backup_id' must be of type str. Given type: {dt}."
        storage_val_err_msg = lambda val: (
            f"'backend' must have one of these values: {STORAGE_NAMES}. " f"Given value: {val}."
        )
        include_type_err_msg = lambda dt: (
            "'include_classes' must be of type str, list of str or None. " f"Given type: {dt}."
        )
        exclude_type_err_msg = lambda dt: (
            "'exclude_classes' must be of type str, list of str or None. " f"Given type: {dt}."
        )
        wait_type_err_msg = lambda dt: (
            f"'wait_for_completion' must be of type bool. Given type: {dt}."
        )
        include_exclude_err_msg = (
            "Either 'include_classes' OR 'exclude_classes' can be set, not both."
        )
        connection_err_msg = "Backup creation failed due to connection error."
        status_code_err_msg = "Backup creation"
        backup_failed_err_msg = lambda status: f"Backup failed: {status}"

        # invalid calls

        with self.assertRaises(TypeError) as error:
            Backup(None).create(
                backup_id=b"My-bucket",
                backend="s3",
            )
        check_error_message(self, error, backup_type_err_msg(bytes))

        with self.assertRaises(ValueError) as error:
            Backup(None).create(
                backup_id="My-bucket",
                backend=b"s3",
            )
        check_error_message(self, error, storage_val_err_msg(b"s3"))

        with self.assertRaises(ValueError) as error:
            Backup(None).create(
                backup_id="My-bucket",
                backend="s4",
            )
        check_error_message(self, error, storage_val_err_msg("s4"))

        with self.assertRaises(TypeError) as error:
            Backup(None).create(
                backup_id="My-bucket",
                backend="s3",
                include_classes=b"MyClass",
            )
        check_error_message(self, error, include_type_err_msg(bytes))

        with self.assertRaises(TypeError) as error:
            Backup(None).create(
                backup_id="My-bucket",
                backend="s3",
                exclude_classes=b"MyClass",
            )
        check_error_message(self, error, exclude_type_err_msg(bytes))

        with self.assertRaises(TypeError) as error:
            Backup(None).create(
                backup_id="My-bucket",
                backend="s3",
                include_classes="MyClass1",
                exclude_classes="MyClass2",
            )
        check_error_message(self, error, include_exclude_err_msg)

        with self.assertRaises(TypeError) as error:
            Backup(None).create(
                backup_id="My-bucket",
                backend="s3",
                include_classes="MyClass1",
                exclude_classes=[],
                wait_for_completion=0,
            )
        check_error_message(self, error, wait_type_err_msg(int))

        mock_conn = mock_connection_func("post", side_effect=RequestsConnectionError)
        with self.assertRaises(RequestsConnectionError) as error:
            Backup(mock_conn).create(
                backup_id="My-bucket",
                backend="s3",
            )
        check_error_message(self, error, connection_err_msg)

        mock_conn = mock_connection_func("post", status_code=404)
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            Backup(mock_conn).create(
                backup_id="My-bucket",
                backend="s3",
            )
        check_startswith_error_message(self, error, status_code_err_msg)

        mock_conn = mock_connection_func(
            "post", status_code=200, return_json={"classes": ["Test"], "error": "Test2"}
        )
        mock_status.return_value = {"status": "FAILED", "error": "test"}
        with self.assertRaises(BackupFailedException) as error:
            Backup(mock_conn).create(
                backup_id="My-bucket",
                backend="s3",
                wait_for_completion=True,
            )
        check_error_message(
            self,
            error,
            backup_failed_err_msg({"classes": ["Test"], "error": "test", "status": "FAILED"}),
        )

        # valid calls
        mock_conn = mock_connection_func("post", status_code=200, return_json={"status": "TEST"})
        mock_status.return_value = {"status": "FAILED", "error": "test"}
        result = Backup(mock_conn).create(
            backup_id="My-Bucket",
            backend="s3",
        )
        self.assertDictEqual(result, {"status": "TEST"})
        mock_conn.post.assert_called_with(
            path="/backups/s3",
            weaviate_object={
                "id": "my-bucket",
                "config": {},
                "include": [],
                "exclude": [],
            },
        )

        mock_conn = mock_connection_func("post", status_code=200, return_json={"status": "TEST"})
        mock_status.return_value = {"status": "FAILED", "error": "test"}
        result = Backup(mock_conn).create(
            backup_id="My-Bucket", backend="gcs", include_classes="myClass"
        )
        self.assertDictEqual(result, {"status": "TEST"})
        mock_conn.post.assert_called_with(
            path="/backups/gcs",
            weaviate_object={
                "id": "my-bucket",
                "config": {},
                "include": ["MyClass"],
                "exclude": [],
            },
        )

        mock_conn = mock_connection_func("post", status_code=200, return_json={"status": "TEST"})
        mock_status.return_value = {"status": "FAILED", "error": "test"}
        result = Backup(mock_conn).create(
            backup_id="My-Bucket", backend="filesystem", include_classes=["class1", "Class2"]
        )
        self.assertDictEqual(result, {"status": "TEST"})
        mock_conn.post.assert_called_with(
            path="/backups/filesystem",
            weaviate_object={
                "id": "my-bucket",
                "config": {},
                "include": ["Class1", "Class2"],
                "exclude": [],
            },
        )

        mock_conn = mock_connection_func("post", status_code=200, return_json={"status": "TEST"})
        mock_status.return_value = {"status": "FAILED", "error": "test"}
        result = Backup(mock_conn).create(
            backup_id="My-Bucket",
            backend="s3",
            exclude_classes="myClass",
            include_classes=None,
        )
        self.assertDictEqual(result, {"status": "TEST"})
        mock_conn.post.assert_called_with(
            path="/backups/s3",
            weaviate_object={
                "id": "my-bucket",
                "config": {},
                "include": [],
                "exclude": ["MyClass"],
            },
        )

        mock_conn = mock_connection_func("post", status_code=200, return_json={"status": "TEST"})
        mock_status.return_value = {"status": "FAILED", "error": "test"}
        result = Backup(mock_conn).create(
            backup_id="My-Bucket",
            backend="s3",
            exclude_classes=["class1", "Class2"],
            include_classes=[],
        )
        self.assertDictEqual(result, {"status": "TEST"})
        mock_conn.post.assert_called_with(
            path="/backups/s3",
            weaviate_object={
                "id": "my-bucket",
                "config": {},
                "include": [],
                "exclude": ["Class1", "Class2"],
            },
        )

        mock_conn = mock_connection_func("post", status_code=200, return_json={"status": "TEST"})
        mock_status.return_value = {"status": "SUCCESS"}
        result = Backup(mock_conn).create(
            backup_id="My-Bucket2",
            backend="s3",
            wait_for_completion=True,
        )
        self.assertDictEqual(result, {"status": "SUCCESS"})
        mock_conn.post.assert_called_with(
            path="/backups/s3",
            weaviate_object={
                "id": "my-bucket2",
                "config": {},
                "include": [],
                "exclude": [],
            },
        )

        with patch("weaviate.backup.backup.sleep") as mock_sleep:

            def override_mock_status():
                mock_status.return_value = {"status": "SUCCESS"}

            mock_sleep.side_effect = lambda n: override_mock_status()
            mock_conn = mock_connection_func(
                "post", status_code=200, return_json={"status": "TEST"}
            )
            mock_status.return_value = {"status": "TRANSFERRING"}
            result = Backup(mock_conn).create(
                backup_id="my-Bucket_2",
                backend="gcs",
                wait_for_completion=True,
            )
            self.assertDictEqual(result, {"status": "SUCCESS"})
            mock_conn.post.assert_called_with(
                path="/backups/gcs",
                weaviate_object={
                    "id": "my-bucket_2",
                    "config": {},
                    "include": [],
                    "exclude": [],
                },
            )

    def test_get_create_status(self):
        """
        Test the `get_create_status` method.
        """

        # error messages
        backup_type_err_msg = lambda dt: f"'backup_id' must be of type str. Given type: {dt}."
        storage_val_err_msg = lambda val: (
            f"'backend' must have one of these values: {STORAGE_NAMES}. " f"Given value: {val}."
        )
        connection_err_msg = "Backup creation status failed due to connection error."
        status_code_err_msg = "Backup status check"

        # invalid calls
        with self.assertRaises(TypeError) as error:
            Backup(None).get_create_status(
                backup_id=b"My-bucket",
                backend="s3",
            )
        check_error_message(self, error, backup_type_err_msg(bytes))

        with self.assertRaises(ValueError) as error:
            Backup(None).get_create_status(
                backup_id="My-bucket",
                backend=b"s3",
            )
        check_error_message(self, error, storage_val_err_msg(b"s3"))

        with self.assertRaises(ValueError) as error:
            Backup(None).get_create_status(
                backup_id="My-bucket",
                backend="s4",
            )
        check_error_message(self, error, storage_val_err_msg("s4"))

        mock_conn = mock_connection_func("get", side_effect=RequestsConnectionError)
        with self.assertRaises(RequestsConnectionError) as error:
            Backup(mock_conn).get_create_status(
                backup_id="My-bucket",
                backend="s3",
            )
        check_error_message(self, error, connection_err_msg)

        mock_conn = mock_connection_func("get", status_code=404)
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            Backup(mock_conn).get_create_status(
                backup_id="My-bucket",
                backend="s3",
            )
        check_startswith_error_message(self, error, status_code_err_msg)

        # valid calls

        mock_conn = mock_connection_func("get", status_code=200, return_json={"status": "TEST"})
        result = Backup(mock_conn).get_create_status(
            backup_id="My-Bucket",
            backend="s3",
        )
        self.assertDictEqual(result, {"status": "TEST"})
        mock_conn.get.assert_called_with(
            path="/backups/s3/my-bucket",
        )

        mock_conn = mock_connection_func(
            "get", status_code=200, return_json={"status": "TEST", "error": None}
        )
        result = Backup(mock_conn).get_create_status(
            backup_id="My-Bucket",
            backend="gcs",
        )
        self.assertDictEqual(result, {"status": "TEST", "error": None})
        mock_conn.get.assert_called_with(
            path="/backups/gcs/my-bucket",
        )

        mock_conn = mock_connection_func("get", status_code=200, return_json={"status": "TEST2"})
        result = Backup(mock_conn).get_create_status(
            backup_id="My-Bucket123",
            backend="filesystem",
        )
        self.assertDictEqual(result, {"status": "TEST2"})
        mock_conn.get.assert_called_with(
            path="/backups/filesystem/my-bucket123",
        )

    @patch("weaviate.backup.backup.Backup.get_restore_status")
    def test_restore(self, mock_status):
        """
        Test the `create` method.
        """

        # error messages
        backup_type_err_msg = lambda dt: f"'backup_id' must be of type str. Given type: {dt}."
        storage_val_err_msg = lambda val: (
            f"'backend' must have one of these values: {STORAGE_NAMES}. " f"Given value: {val}."
        )
        include_type_err_msg = lambda dt: (
            "'include_classes' must be of type str, list of str or None. " f"Given type: {dt}."
        )
        exclude_type_err_msg = lambda dt: (
            "'exclude_classes' must be of type str, list of str or None. " f"Given type: {dt}."
        )
        wait_type_err_msg = lambda dt: (
            f"'wait_for_completion' must be of type bool. Given type: {dt}."
        )
        include_exclude_err_msg = (
            "Either 'include_classes' OR 'exclude_classes' can be set, not both."
        )
        connection_err_msg = "Backup restore failed due to connection error."
        status_code_err_msg = "Backup restore"
        backup_failed_err_msg = lambda status: f"Backup restore failed: {status}"

        # invalid calls

        with self.assertRaises(TypeError) as error:
            Backup(None).restore(
                backup_id=b"My-bucket",
                backend="s3",
            )
        check_error_message(self, error, backup_type_err_msg(bytes))

        with self.assertRaises(ValueError) as error:
            Backup(None).restore(
                backup_id="My-bucket",
                backend=b"s3",
            )
        check_error_message(self, error, storage_val_err_msg(b"s3"))

        with self.assertRaises(ValueError) as error:
            Backup(None).restore(
                backup_id="My-bucket",
                backend="s4",
            )
        check_error_message(self, error, storage_val_err_msg("s4"))

        with self.assertRaises(TypeError) as error:
            Backup(None).restore(
                backup_id="My-bucket",
                backend="s3",
                include_classes=b"MyClass",
            )
        check_error_message(self, error, include_type_err_msg(bytes))

        with self.assertRaises(TypeError) as error:
            Backup(None).restore(
                backup_id="My-bucket",
                backend="s3",
                exclude_classes=b"MyClass",
            )
        check_error_message(self, error, exclude_type_err_msg(bytes))

        with self.assertRaises(TypeError) as error:
            Backup(None).restore(
                backup_id="My-bucket",
                backend="s3",
                include_classes="MyClass1",
                exclude_classes="MyClass2",
            )
        check_error_message(self, error, include_exclude_err_msg)

        with self.assertRaises(TypeError) as error:
            Backup(None).restore(
                backup_id="My-bucket",
                backend="s3",
                include_classes="MyClass1",
                exclude_classes=[],
                wait_for_completion=0,
            )
        check_error_message(self, error, wait_type_err_msg(int))

        mock_conn = mock_connection_func("post", side_effect=RequestsConnectionError)
        with self.assertRaises(RequestsConnectionError) as error:
            Backup(mock_conn).restore(
                backup_id="My-bucket",
                backend="s3",
            )
        check_error_message(self, error, connection_err_msg)

        mock_conn = mock_connection_func("post", status_code=404)
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            Backup(mock_conn).restore(
                backup_id="My-bucket",
                backend="s3",
            )
        check_startswith_error_message(self, error, status_code_err_msg)

        mock_conn = mock_connection_func(
            "post", status_code=200, return_json={"classes": ["Test"], "error": "Test2"}
        )
        mock_status.return_value = {"status": "FAILED", "error": "test"}
        with self.assertRaises(BackupFailedException) as error:
            Backup(mock_conn).restore(
                backup_id="My-bucket",
                backend="s3",
                wait_for_completion=True,
            )
        check_error_message(
            self,
            error,
            backup_failed_err_msg({"classes": ["Test"], "error": "test", "status": "FAILED"}),
        )

        # valid calls
        mock_conn = mock_connection_func("post", status_code=200, return_json={"status": "TEST"})
        mock_status.return_value = {"status": "FAILED", "error": "test"}
        result = Backup(mock_conn).restore(
            backup_id="My-Bucket",
            backend="s3",
        )
        self.assertDictEqual(result, {"status": "TEST"})
        mock_conn.post.assert_called_with(
            path="/backups/s3/my-bucket/restore",
            weaviate_object={
                "config": {},
                "include": [],
                "exclude": [],
            },
        )

        mock_conn = mock_connection_func("post", status_code=200, return_json={"status": "TEST"})
        mock_status.return_value = {"status": "FAILED", "error": "test"}
        result = Backup(mock_conn).restore(
            backup_id="My-Bucket", backend="gcs", include_classes="myClass"
        )
        self.assertDictEqual(result, {"status": "TEST"})
        mock_conn.post.assert_called_with(
            path="/backups/gcs/my-bucket/restore",
            weaviate_object={
                "config": {},
                "include": ["MyClass"],
                "exclude": [],
            },
        )

        mock_conn = mock_connection_func("post", status_code=200, return_json={"status": "TEST"})
        mock_status.return_value = {"status": "FAILED", "error": "test"}
        result = Backup(mock_conn).restore(
            backup_id="My-Bucket", backend="filesystem", include_classes=["class1", "Class2"]
        )
        self.assertDictEqual(result, {"status": "TEST"})
        mock_conn.post.assert_called_with(
            path="/backups/filesystem/my-bucket/restore",
            weaviate_object={
                "config": {},
                "include": ["Class1", "Class2"],
                "exclude": [],
            },
        )

        mock_conn = mock_connection_func("post", status_code=200, return_json={"status": "TEST"})
        mock_status.return_value = {"status": "FAILED", "error": "test"}
        result = Backup(mock_conn).restore(
            backup_id="My-Bucket",
            backend="s3",
            exclude_classes="myClass",
            include_classes=None,
        )
        self.assertDictEqual(result, {"status": "TEST"})
        mock_conn.post.assert_called_with(
            path="/backups/s3/my-bucket/restore",
            weaviate_object={
                "config": {},
                "include": [],
                "exclude": ["MyClass"],
            },
        )

        mock_conn = mock_connection_func("post", status_code=200, return_json={"status": "TEST"})
        mock_status.return_value = {"status": "FAILED", "error": "test"}
        result = Backup(mock_conn).restore(
            backup_id="My-Bucket",
            backend="s3",
            exclude_classes=["class1", "Class2"],
            include_classes=[],
        )
        self.assertDictEqual(result, {"status": "TEST"})
        mock_conn.post.assert_called_with(
            path="/backups/s3/my-bucket/restore",
            weaviate_object={
                "config": {},
                "include": [],
                "exclude": ["Class1", "Class2"],
            },
        )

        mock_conn = mock_connection_func("post", status_code=200, return_json={"status": "TEST"})
        mock_status.return_value = {"status": "SUCCESS"}
        result = Backup(mock_conn).restore(
            backup_id="My-Bucket2",
            backend="s3",
            wait_for_completion=True,
        )
        self.assertDictEqual(result, {"status": "SUCCESS"})
        mock_conn.post.assert_called_with(
            path="/backups/s3/my-bucket2/restore",
            weaviate_object={
                "config": {},
                "include": [],
                "exclude": [],
            },
        )

        with patch("weaviate.backup.backup.sleep") as mock_sleep:

            def override_mock_status():
                mock_status.return_value = {"status": "SUCCESS"}

            mock_sleep.side_effect = lambda n: override_mock_status()
            mock_conn = mock_connection_func(
                "post", status_code=200, return_json={"status": "TEST"}
            )
            mock_status.return_value = {"status": "TRANSFERRING"}
            result = Backup(mock_conn).restore(
                backup_id="my-Bucket_2",
                backend="gcs",
                wait_for_completion=True,
            )
            self.assertDictEqual(result, {"status": "SUCCESS"})
            mock_conn.post.assert_called_with(
                path="/backups/gcs/my-bucket_2/restore",
                weaviate_object={
                    "config": {},
                    "include": [],
                    "exclude": [],
                },
            )

    def test_get_restore_status(self):
        """
        Test the `get_restore_status` method.
        """

        # error messages
        backup_type_err_msg = lambda dt: f"'backup_id' must be of type str. Given type: {dt}."
        storage_val_err_msg = lambda val: (
            f"'backend' must have one of these values: {STORAGE_NAMES}. " f"Given value: {val}."
        )
        connection_err_msg = "Backup restore status failed due to connection error."
        status_code_err_msg = "Backup restore status check"

        # invalid calls
        with self.assertRaises(TypeError) as error:
            Backup(None).get_restore_status(
                backup_id=b"My-bucket",
                backend="s3",
            )
        check_error_message(self, error, backup_type_err_msg(bytes))

        with self.assertRaises(ValueError) as error:
            Backup(None).get_restore_status(
                backup_id="My-bucket",
                backend=b"s3",
            )
        check_error_message(self, error, storage_val_err_msg(b"s3"))

        with self.assertRaises(ValueError) as error:
            Backup(None).get_restore_status(
                backup_id="My-bucket",
                backend="s4",
            )
        check_error_message(self, error, storage_val_err_msg("s4"))

        mock_conn = mock_connection_func("get", side_effect=RequestsConnectionError)
        with self.assertRaises(RequestsConnectionError) as error:
            Backup(mock_conn).get_restore_status(
                backup_id="My-bucket",
                backend="s3",
            )
        check_error_message(self, error, connection_err_msg)

        mock_conn = mock_connection_func("get", status_code=404)
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            Backup(mock_conn).get_restore_status(
                backup_id="My-bucket",
                backend="s3",
            )
        check_startswith_error_message(self, error, status_code_err_msg)

        # valid calls

        mock_conn = mock_connection_func("get", status_code=200, return_json={"status": "TEST"})
        result = Backup(mock_conn).get_restore_status(
            backup_id="My-Bucket",
            backend="s3",
        )
        self.assertDictEqual(result, {"status": "TEST"})
        mock_conn.get.assert_called_with(
            path="/backups/s3/my-bucket/restore",
        )

        mock_conn = mock_connection_func(
            "get", status_code=200, return_json={"status": "TEST", "error": None}
        )
        result = Backup(mock_conn).get_restore_status(
            backup_id="My-Bucket",
            backend="gcs",
        )
        self.assertDictEqual(result, {"status": "TEST", "error": None})
        mock_conn.get.assert_called_with(
            path="/backups/gcs/my-bucket/restore",
        )

        mock_conn = mock_connection_func("get", status_code=200, return_json={"status": "TEST2"})
        result = Backup(mock_conn).get_restore_status(
            backup_id="My-Bucket123",
            backend="filesystem",
        )
        self.assertDictEqual(result, {"status": "TEST2"})
        mock_conn.get.assert_called_with(
            path="/backups/filesystem/my-bucket123/restore",
        )



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import contextlib
import os
import sys

from pkg_resources import DistributionNotFound, get_distribution

# -- Path setup --------------------------------------------------------------
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

sys.path.insert(0, os.path.abspath(".."))


@contextlib.contextmanager
def chdir(directory):
    curdir = os.curdir
    try:
        os.chdir(directory)
        yield
    finally:
        os.chdir(curdir)


try:
    dist = get_distribution("weaviate")
except DistributionNotFound:
    # The project is not installed in readthedocs environment. Read the version with setuptools_scm.
    import setuptools_scm

    with chdir(".."):
        release = setuptools_scm.get_version()
else:
    release = dist.version

# -- Project information -----------------------------------------------------

project = "Weaviate Python Client"
copyright = "2021-2023, Weaviate"
author = "Weaviate"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
]

# Make sure the target is unique
autosectionlabel_prefix_document = True

autoclass_content = "both"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'classic'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []



from typing import Generic, Optional, Type, Union, cast, overload
from typing_extensions import is_typeddict

from weaviate.collections.classes.config import (
    ConsistencyLevel,
)
from weaviate.collections.classes.grpc import MetadataQuery, PROPERTIES
from weaviate.collections.classes.types import Properties, TProperties
from weaviate.collections.base import _CollectionBase
from weaviate.collections.aggregate import _AggregateCollection, _AggregateGroupByCollection
from weaviate.collections.config import _ConfigCollection
from weaviate.collections.data import _DataCollection
from weaviate.collections.query import _GenerateCollection, _GroupByCollection, _QueryCollection
from weaviate.collections.iterator import _ObjectIterator
from weaviate.collections.tenants import _Tenants
from weaviate.connect import Connection


class _Collection(_CollectionBase, Generic[Properties]):
    def __init__(
        self,
        connection: Connection,
        name: str,
        consistency_level: Optional[ConsistencyLevel] = None,
        tenant: Optional[str] = None,
        type_: Optional[Type[Properties]] = None,
    ) -> None:
        super().__init__(name)
        self._connection = connection

        self.aggregate = _AggregateCollection(
            self._connection, self.name, consistency_level, tenant
        )
        """This namespace includes all the querying methods available to you when using Weaviate's standard aggregation capabilities."""
        self.aggregate_group_by = _AggregateGroupByCollection(
            self._connection, self.name, consistency_level, tenant
        )
        """This namespace includes all the aggregate methods available to you when using Weaviate's aggregation group-by capabilities."""
        self.config = _ConfigCollection(self._connection, self.name, tenant)
        """This namespace includes all the CRUD methods available to you when modifying the configuration of the collection in Weaviate."""
        self.data = _DataCollection[Properties](
            connection, self.name, consistency_level, tenant, type_
        )
        """This namespace includes all the CUD methods available to you when modifying the data of the collection in Weaviate."""
        self.generate = _GenerateCollection(connection, self.name, consistency_level, tenant, type_)
        """This namespace includes all the querying methods available to you when using Weaviate's generative capabilities."""
        self.query_group_by = _GroupByCollection(
            connection, self.name, consistency_level, tenant, type_
        )
        """This namespace includes all the querying methods available to you when using Weaviate's querying group-by capabilities."""
        self.query = _QueryCollection[Properties](
            connection, self.name, self.data, consistency_level, tenant, type_
        )
        """This namespace includes all the querying methods available to you when using Weaviate's standard query capabilities."""
        self.tenants = _Tenants(connection, self.name)
        """This namespace includes all the CRUD methods available to you when modifying the tenants of a multi-tenancy-enabled collection in Weaviate."""

        self.__tenant = tenant
        self.__consistency_level = consistency_level
        self.__type = type_

    def with_tenant(self, tenant: Optional[str] = None) -> "_Collection":
        """Use this method to return a collection object specific to a single tenant.

        If multi-tenancy is not configured for this collection then Weaviate will throw an error.

        Arguments:
            `tenant`
                The name of the tenant to use.
        """
        return _Collection(self._connection, self.name, self.__consistency_level, tenant)

    def with_consistency_level(
        self, consistency_level: Optional[ConsistencyLevel] = None
    ) -> "_Collection":
        """Use this method to return a collection object specific to a single consistency level.

        If replication is not configured for this collection then Weaviate will throw an error.

        Arguments:
            `consistency_level`
                The consistency level to use.
        """
        return _Collection(self._connection, self.name, consistency_level, self.__tenant)

    def __len__(self) -> int:
        total = self.aggregate.over_all(total_count=True).total_count
        assert total is not None
        return total

    @overload
    def iterator(
        self,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> _ObjectIterator[Properties]:
        ...

    @overload
    def iterator(
        self,
        return_metadata: Optional[MetadataQuery] = None,
        *,
        return_properties: Type[TProperties],
    ) -> _ObjectIterator[TProperties]:
        ...

    def iterator(
        self,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[Union[PROPERTIES, Type[TProperties]]] = None,
    ) -> Union[_ObjectIterator[Properties], _ObjectIterator[TProperties]]:
        """Use this method to return an iterator over the objects in the collection.

        This iterator keeps a record of the last object that it returned to be used in each subsequent call to
        Weaviate. Once the collection is exhausted, the iterator exits.

        If `return_metadata` and `return_properties` are not provided, all the data of each object will be
        requested from Weaviate except for its vector as this is an expensive operation. Specify `return_metadata`
        and `return_properties` to only request the data that you need.

        Arguments:
            `return_metadata`
                The metadata to return with each object.
            `return_properties`
                The properties to return with each object.

        Raises:
            `requests.ConnectionError`
                If the network connection to Weaviate fails.
            `weaviate.UnexpectedStatusCodeException`
                If Weaviate reports a non-OK status.
        """
        if is_typeddict(return_properties):
            return_properties = cast(Type[TProperties], return_properties)
            return _ObjectIterator[TProperties](
                lambda limit, alpha, meta: self.query.fetch_objects(  # type: ignore # shouldn't be needed but mypy complains
                    limit=limit,
                    after=alpha,
                    return_metadata=meta,
                    return_properties=return_properties,
                ).objects,
                return_metadata,
                return_properties,
            )
        if return_properties is None and self.__type is not None:
            _type = cast(Type[Properties], self.__type)
            return _ObjectIterator[Properties](
                lambda limit, alpha, meta: self.query.fetch_objects(
                    limit=limit,
                    after=alpha,
                    return_metadata=meta,
                    return_properties=_type,
                ).objects,
                return_metadata,
                _type,
            )
        props = cast(PROPERTIES, return_properties)
        return _ObjectIterator[Properties](
            lambda limit, alpha, meta: self.query.fetch_objects(
                limit=limit,
                after=alpha,
                return_metadata=meta,
                return_properties=props,
            ).objects,
            return_metadata,
            props,
        )



import json

import pytest
from pytest_httpserver import HTTPServer, HeaderValueMatcher
from werkzeug.wrappers import Response

from weaviate.connect.connection import ConnectionParams, ProtocolParams

MOCK_IP = "127.0.0.1"
MOCK_PORT = 23536
CLIENT_ID = "DoesNotMatter"
MOCK_SERVER_URL = "http://" + MOCK_IP + ":" + str(MOCK_PORT)
MOCK_SERVER_CONNECTION_PARAMS = ConnectionParams(
    http=ProtocolParams(host=MOCK_IP, port=MOCK_PORT, secure=False)
)

# pytest_httpserver 'Authorization' HeaderValueMatcher does not work with Bearer tokens.
# Hence, overwrite it with the default header value matcher that just compares for equality.
HeaderValueMatcher.DEFAULT_MATCHERS[
    "Authorization"
] = HeaderValueMatcher.default_header_value_matcher


@pytest.fixture(scope="session")
def httpserver_listen_address():
    return MOCK_IP, MOCK_PORT


@pytest.fixture(scope="function")
def ready_mock(httpserver: HTTPServer):
    httpserver.expect_request("/v1/.well-known/ready").respond_with_json({})
    yield httpserver


@pytest.fixture(scope="function")
def weaviate_mock(ready_mock):
    ready_mock.expect_request("/v1/meta").respond_with_json({"version": "1.16"})
    ready_mock.expect_request("/v1/nodes").respond_with_json({"nodes": [{"gitHash": "ABC"}]})

    yield ready_mock


@pytest.fixture(scope="function")
def weaviate_no_auth_mock(ready_mock):
    ready_mock.expect_request("/v1/meta").respond_with_json({"version": "1.16"})
    ready_mock.expect_request("/v1/.well-known/openid-configuration").respond_with_response(
        Response(json.dumps({}), status=404)
    )

    yield ready_mock


@pytest.fixture(scope="function")
def weaviate_auth_mock(weaviate_mock):
    weaviate_mock.expect_request("/v1/.well-known/openid-configuration").respond_with_json(
        {
            "href": MOCK_SERVER_URL + "/endpoints",
            "clientId": CLIENT_ID,
        }
    )
    weaviate_mock.expect_request("/endpoints").respond_with_json(
        {"token_endpoint": MOCK_SERVER_URL + "/auth"}
    )
    yield weaviate_mock



"""
Schema class definition.
"""
from dataclasses import dataclass
from enum import Enum
from typing import Any, Union, Optional, List, Dict, cast

from requests.exceptions import ConnectionError as RequestsConnectionError

from weaviate.connect import Connection
from weaviate.exceptions import UnexpectedStatusCodeException
from weaviate.schema.properties import Property
from weaviate.util import (
    _get_dict_from_object,
    _is_sub_schema,
    _capitalize_first_letter,
    _decode_json_response_dict,
    _decode_json_response_list,
)

CLASS_KEYS = {
    "class",
    "vectorIndexType",
    "vectorIndexConfig",
    "moduleConfig",
    "description",
    "vectorizer",
    "properties",
    "invertedIndexConfig",
    "shardingConfig",
    "replicationConfig",
    "multiTenancyConfig",
}

PROPERTY_KEYS = {
    "dataType",
    "name",
    "moduleConfig",
    "description",
    "indexInverted",
    "tokenization",
    "indexFilterable",
    "indexSearchable",
}

_PRIMITIVE_WEAVIATE_TYPES_SET = {
    "string",
    "string[]",
    "int",
    "int[]",
    "boolean",
    "boolean[]",
    "number",
    "number[]",
    "date",
    "date[]",
    "text",
    "text[]",
    "geoCoordinates",
    "blob",
    "phoneNumber",
    "uuid",
    "uuid[]",
    "object",
    "object[]",
}


class TenantActivityStatus(str, Enum):
    """
    TenantActivityStatus class used to describe the activity status of a tenant in Weaviate.

    Attributes
    ----------
    HOT: The tenant is fully active and can be used.
    COLD: The tenant is not active, files stored locally.
    """

    HOT = "HOT"
    COLD = "COLD"


@dataclass
class Tenant:
    """
    Tenant class used to describe a tenant in Weaviate.

    Attributes
    ----------
    activity_status : TenantActivityStatus, optional
        default: "HOT"
    name: the name of the tenant.
    """

    name: str
    activity_status: TenantActivityStatus = TenantActivityStatus.HOT

    def _to_weaviate_object(self) -> Dict[str, str]:
        return {
            "activityStatus": self.activity_status,
            "name": self.name,
        }

    @classmethod
    def _from_weaviate_object(cls, weaviate_object: Dict[str, Any]) -> "Tenant":
        return cls(
            name=weaviate_object["name"],
            activity_status=TenantActivityStatus(weaviate_object.get("activityStatus", "HOT")),
        )


class Schema:
    """
    Schema class used to interact and manipulate schemas or classes.

    Attributes
    ----------
    property : weaviate.schema.properties.Property
        A Property object to create new schema property/ies.
    """

    def __init__(self, connection: Connection):
        """
        Initialize a Schema class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running Weaviate instance.
        """

        self._connection = connection
        self.property = Property(self._connection)

    def create(self, schema: Union[dict, str]) -> None:
        """
        Create the schema of the Weaviate instance, with all classes at once.

        Parameters
        ----------
        schema : dict or str
            Schema as a Python dict, or the path to a JSON file, or the URL of a JSON file.

        Examples
        --------
        >>> article_class = {
        ...     "class": "Article",
        ...     "description": "An article written by an Author",
        ...     "properties": [
        ...         {
        ...             "name": "title",
        ...             "dataType": ["text"],
        ...             "description": "The title the article",
        ...         },
        ...         {
        ...             "name": "hasAuthors",
        ...             "dataType": ["Author"],
        ...             "description": "Authors this article has",
        ...         }
        ...     ]
        ... }
        >>> author_class = {
        ...     "class": "Author",
        ...     "description": "An Author class to store the author information",
        ...     "properties": [
        ...         {
        ...             "name": "name",
        ...             "dataType": ["text"],
        ...             "description": "The name of the author",
        ...         },
        ...         {
        ...             "name": "wroteArticles",
        ...             "dataType": ["Article"],
        ...             "description": "The articles of the author",
        ...         }
        ...     ]
        ... }
        >>> client.schema.create({"classes": [article_class, author_class]})

        If you have your schema saved in the './schema/my_schema.json' you can create it
        directly from the file.

        >>> client.schema.create('./schema/my_schema.json')

        Raises
        ------
        TypeError
            If the 'schema' is neither a string nor a dict.
        ValueError
            If 'schema' can not be converted into a Weaviate schema.
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a non-OK status.
        weaviate.SchemaValidationException
            If the 'schema' could not be validated against the standard format.
        """

        loaded_schema = _get_dict_from_object(schema)
        self._create_classes_with_primitives(loaded_schema["classes"])
        self._create_complex_properties_from_classes(loaded_schema["classes"])

    def create_class(self, schema_class: Union[dict, str]) -> None:
        """
        Create a single class as part of the schema in Weaviate.

        Parameters
        ----------
        schema_class : dict or str
            Class as a Python dict, or the path to a JSON file, or the URL of a JSON file.

        Examples
        --------
        >>> author_class_schema = {
        ...     "class": "Author",
        ...     "description": "An Author class to store the author information",
        ...     "properties": [
        ...         {
        ...             "name": "name",
        ...             "dataType": ["text"],
        ...             "description": "The name of the author",
        ...         },
        ...         {
        ...             "name": "wroteArticles",
        ...             "dataType": ["Article"],
        ...             "description": "The articles of the author",
        ...         }
        ...     ]
        ... }
        >>> client.schema.create_class(author_class_schema)

        If you have your class schema saved in the './schema/my_schema.json' you can create it
        directly from the file.

        >>> client.schema.create_class('./schema/my_schema.json')

        Raises
        ------
        TypeError
            If the 'schema_class' is neither a string nor a dict.
        ValueError
            If 'schema_class' can not be converted into a Weaviate schema.
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a non-OK status.
        weaviate.SchemaValidationException
            If the 'schema_class' could not be validated against the standard format.
        """

        loaded_schema_class = _get_dict_from_object(schema_class)
        self._create_class_with_primitives(loaded_schema_class)
        self._create_complex_properties_from_class(loaded_schema_class)

    def delete_class(self, class_name: str) -> None:
        """
        Delete a schema class from Weaviate. This deletes all associated data.

        Parameters
        ----------
        class_name : str
            The class that should be deleted from Weaviate.

        Examples
        --------
        >>> client.schema.delete_class('Author')

        Raises
        ------
        TypeError
            If 'class_name' argument not of type str.
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a non-OK status.
        """

        if not isinstance(class_name, str):
            raise TypeError(f"Class name was {type(class_name)} instead of str")

        path = f"/schema/{_capitalize_first_letter(class_name)}"
        try:
            response = self._connection.delete(path=path)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Deletion of class.") from conn_err
        if response.status_code != 200:
            raise UnexpectedStatusCodeException("Delete class from schema", response)

    def delete_all(self) -> None:
        """
        Remove the entire schema from the Weaviate instance and all data associated with it.

        Examples
        --------
        >>> client.schema.delete_all()
        """

        schema = self.get()
        classes = schema.get("classes", [])
        for _class in classes:
            self.delete_class(_class["class"])

    def exists(self, class_name: str) -> bool:
        """
        Check if class exists in Weaviate.

        Parameters
        ----------
        class_name : str
            The class whose existence is being checked.

        Examples
        --------
        >>> client.schema.exists(class_name="Exists")
        True

        >>> client.schema.exists(class_name="DoesNotExists")
        False

        Returns
        -------
        bool
            True if the class exists,
            False otherwise.
        """

        if not isinstance(class_name, str):
            raise TypeError(
                f"'class_name' argument must be of type `str`! Given type: {type(class_name)}."
            )

        path = f"/schema/{_capitalize_first_letter(class_name)}"

        try:
            response = self._connection.get(path=path)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                "Checking class existence could not be done."
            ) from conn_err
        if response.status_code == 200:
            return True
        elif response.status_code == 404:
            return False

        raise UnexpectedStatusCodeException("Check if class exists", response)

    def contains(self, schema: Optional[Union[dict, str]] = None) -> bool:
        """
        Check if Weaviate already contains a schema.

        Parameters
        ----------
        schema : dict or str, optional
            Schema as a Python dict, or the path to a JSON file or the URL of a JSON file.
            If a schema is given it is checked if this specific schema is already loaded.
            It will test only this schema. If the given schema is a subset of the loaded
            schema it will still return true, by default None.

        Examples
        --------
        >>> schema = client.schema.get()
        >>> client.schema.contains(schema)
        True
        >>> schema = client.schema.get()
        >>> schema['classes'].append(
            {
                "class": "Animal",
                "description": "An Animal",
                "properties": [
                    {
                        "name": "type",
                        "dataType": ["text"],
                        "description": "The animal type",
                    }
                ]
            }
        )
        >>> client.schema.contains(schema)
        False

        Returns
        -------
        bool
            True if a schema is present,
            False otherwise.
        """

        loaded_schema = self.get()

        if schema is not None:
            sub_schema = _get_dict_from_object(schema)
            return _is_sub_schema(sub_schema, loaded_schema)

        if len(loaded_schema["classes"]) == 0:
            return False
        return True

    def update_config(self, class_name: str, config: dict) -> None:
        """
        Update a schema configuration for a specific class.

        Parameters
        ----------
        class_name : str
            The class for which to update the schema configuration.
        config : dict
            The configurations to update (MUST follow schema format).

        Example
        -------
        In the example below we have a Weaviate instance with a class 'Test'.

        >>> client.schema.get('Test')
        {
            'class': 'Test',
            ...
            'vectorIndexConfig': {
                'ef': -1,
                ...
            },
            ...
        }
        >>> client.schema.update_config(
        ...     class_name='Test',
        ...     config={
        ...         'vectorIndexConfig': {
        ...             'ef': 100,
        ...         }
        ...     }
        ... )
        >>> client.schema.get('Test')
        {
            'class': 'Test',
            ...
            'vectorIndexConfig': {
                'ef': 100,
                ...
            },
            ...
        }

        NOTE: When updating schema configuration, the 'config' MUST be sub-set of the schema,
        starting at the top level. In the example above we update 'ef' value, and for this we
        included the 'vectorIndexConfig' top level too.

        Raises
        ------
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a non-OK status.
        """

        class_name = _capitalize_first_letter(class_name)
        class_schema = self.get(class_name)
        new_class_schema = _update_nested_dict(class_schema, config)

        path = "/schema/" + class_name
        try:
            response = self._connection.put(path=path, weaviate_object=new_class_schema)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                "Class schema configuration could not be updated."
            ) from conn_err
        if response.status_code != 200:
            raise UnexpectedStatusCodeException("Update class schema configuration", response)

    def get(self, class_name: Optional[str] = None) -> dict:
        """
        Get the schema from Weaviate.

        Parameters
        ----------
        class_name : str, optional
            The class for which to return the schema. If NOT provided the whole schema is returned,
            otherwise only the schema of this class is returned. By default None.

        Returns
        -------
        dict
            A dict containing the schema. The schema may be empty.
            To see if a schema has already been loaded, use the `contains` method.

        Examples
        --------
        No schema present in client

        >>> client.schema.get()
        {'classes': []}

        Schema present in client

        >>> client.schema.get()
        {
            "classes": [
                {
                "class": "Animal",
                "description": "An Animal",
                "invertedIndexConfig": {
                    "cleanupIntervalSeconds": 60
                },
                "properties": [
                    {
                    "dataType": ["text"],
                    "description": "The animal type",
                    "name": "type"
                    }
                ],
                "vectorIndexConfig": {
                    "cleanupIntervalSeconds": 300,
                    "maxConnections": 64,
                    "efConstruction": 128,
                    "vectorCacheMaxObjects": 500000
                },
                "vectorIndexType": "hnsw",
                "vectorizer": "text2vec-contextionary",
                "replicationConfig": {
                    "factor": 1
                }
                }
            ]
        }

        >>> client.schema.get('Animal')
        {
            "class": "Animal",
            "description": "An Animal",
            "invertedIndexConfig": {
                "cleanupIntervalSeconds": 60
            },
            "properties": [
                {
                "dataType": ["text"],
                "description": "The animal type",
                "name": "type"
                }
            ],
            "vectorIndexConfig": {
                "cleanupIntervalSeconds": 300,
                "maxConnections": 64,
                "efConstruction": 128,
                "vectorCacheMaxObjects": 500000
            },
            "vectorIndexType": "hnsw",
            "vectorizer": "text2vec-contextionary",
            "replicationConfig": {
                "factor": 1
            }
        }

        Raises
        ------
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a non-OK status.
        """

        path = "/schema"
        if class_name is not None:
            if not isinstance(class_name, str):
                raise TypeError(
                    "'class_name' argument must be of type `str`! "
                    f"Given type: {type(class_name)}"
                )
            path = f"/schema/{_capitalize_first_letter(class_name)}"

        try:
            response = self._connection.get(path=path)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Schema could not be retrieved.") from conn_err

        res = _decode_json_response_dict(response, "Get schema")
        assert res is not None
        return res

    def get_class_shards(self, class_name: str) -> list:
        """
        Get the status of all shards in an index.

        Parameters
        ----------
        class_name : str
            The class for which to return the status of all shards in an index.

        Returns
        -------
        list
            The list of shards configuration.

        Examples
        --------
        Schema contains a single class: Article

        >>> client.schema.get_class_shards('Article')
        [{'name': '2rPgsA2yngW3', 'status': 'READY'}]

        Raises
        ------
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a non-OK status.
        """

        if not isinstance(class_name, str):
            raise TypeError(
                "'class_name' argument must be of type `str`! " f"Given type: {type(class_name)}."
            )
        path = f"/schema/{_capitalize_first_letter(class_name)}/shards"

        try:
            response = self._connection.get(path=path)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                "Class shards' status could not be retrieved due to connection error."
            ) from conn_err

        res = _decode_json_response_list(response, "Get shards' status")
        assert res is not None
        return res

    def update_class_shard(
        self,
        class_name: str,
        status: str,
        shard_name: Optional[str] = None,
    ) -> list:
        """
        Get the status of all shards in an index.

        Parameters
        ----------
        class_name : str
            The class for which to update the status of all shards in an index.
        status : str
            The new status of the shard. The available options are: 'READY' and 'READONLY'.
        shard_name : str or None, optional
            The shard name for which to update the status of the class of the shard. If None then
            all the shards are going to be updated to the 'status'. By default None.

        Returns
        -------
        list
            The updated statuses.

        Examples
        --------
        Schema contains a single class: Article

        >>> client.schema.get_class_shards('Article')
        [{'name': 'node1', 'status': 'READY'}, {'name': 'node2', 'status': 'READY'}]

        For a specific shard:

        >>> client.schema.update_class_shard('Article', 'READONLY', 'node2')
        {'status': 'READONLY'}
        >>> client.schema.get_class_shards('Article')
        [{'name': 'node1', 'status': 'READY'}, {'name': 'node2', 'status': 'READONLY'}]

        For all shards of the class:

        >>> client.schema.update_class_shard('Article', 'READONLY')
        [{'status': 'READONLY'},{'status': 'READONLY'}]
        >>> client.schema.get_class_shards('Article')
        [{'name': 'node1', 'status': 'READONLY'}, {'name': 'node2', 'status': 'READONLY'}]


        Raises
        ------
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a non-OK status.
        """

        if not isinstance(class_name, str):
            raise TypeError(
                "'class_name' argument must be of type `str`! " f"Given type: {type(class_name)}."
            )
        if not isinstance(shard_name, str) and shard_name is not None:
            raise TypeError(
                "'shard_name' argument must be of type `str`! " f"Given type: {type(shard_name)}."
            )
        if not isinstance(status, str):
            raise TypeError(
                "'status' argument must be of type `str`! " f"Given type: {type(status)}."
            )

        if shard_name is None:
            shards_config = self.get_class_shards(
                class_name=class_name,
            )
            shard_names = [shard_config["name"] for shard_config in shards_config]
        else:
            shard_names = [shard_name]

        data = {"status": status}

        to_return = []

        for _shard_name in shard_names:
            path = f"/schema/{_capitalize_first_letter(class_name)}/shards/{_shard_name}"
            try:
                response = self._connection.put(
                    path=path,
                    weaviate_object=data,
                )
            except RequestsConnectionError as conn_err:
                raise RequestsConnectionError(
                    f"Class shards' status could not be updated for shard '{_shard_name}' due to "
                    "connection error."
                ) from conn_err

            to_return.append(
                _decode_json_response_dict(response, f"Update shard '{_shard_name}' status")
            )

        if shard_name is None:
            return to_return
        return cast(list, to_return[0])

    def _create_complex_properties_from_class(self, schema_class: dict) -> None:
        """
        Add cross-references to an already existing class.

        Parameters
        ----------
        schema_class : dict
            Description of the class that should be added.

        Raises
        ------
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a non-OK status.
        """

        if "properties" not in schema_class:
            # Class has no properties - nothing to do
            return
        for property_ in schema_class["properties"]:
            if _property_is_primitive(property_["dataType"]):
                continue

            # Create the property object. All complex dataTypes should be capitalized.
            schema_property = {
                "dataType": [_capitalize_first_letter(dtype) for dtype in property_["dataType"]],
                "name": property_["name"],
            }

            for property_field in PROPERTY_KEYS - {"name", "dataType"}:
                if property_field in property_:
                    schema_property[property_field] = property_[property_field]

            path = "/schema/" + _capitalize_first_letter(schema_class["class"]) + "/properties"
            try:
                response = self._connection.post(path=path, weaviate_object=schema_property)
            except RequestsConnectionError as conn_err:
                raise RequestsConnectionError(
                    "Property may not have been created properly."
                ) from conn_err
            if response.status_code != 200:
                raise UnexpectedStatusCodeException("Add properties to classes", response)

    def _create_complex_properties_from_classes(self, schema_classes_list: list) -> None:
        """
        Add cross-references to already existing classes.

        Parameters
        ----------
        schema_classes_list : list
            A list of classes as they are found in a schema JSON description.
        """

        for schema_class in schema_classes_list:
            self._create_complex_properties_from_class(schema_class)

    def _create_class_with_primitives(self, weaviate_class: dict) -> None:
        """
        Create class with only primitives.

        Parameters
        ----------
        weaviate_class : dict
            A single Weaviate formatted class

        Raises
        ------
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a non-OK status.
        """

        # Create the class
        schema_class = {
            "class": _capitalize_first_letter(weaviate_class["class"]),
            "properties": [],
        }

        for class_field in CLASS_KEYS - {"class", "properties"}:
            if class_field in weaviate_class:
                schema_class[class_field] = weaviate_class[class_field]

        if "properties" in weaviate_class:
            schema_class["properties"] = _get_primitive_properties(weaviate_class["properties"])

        # Add the item
        try:
            response = self._connection.post(path="/schema", weaviate_object=schema_class)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Class may not have been created properly.") from conn_err
        if response.status_code != 200:
            raise UnexpectedStatusCodeException("Create class", response)

    def _create_classes_with_primitives(self, schema_classes_list: list) -> None:
        """
        Create all the classes in the list and primitive properties.
        This function does not create references,
        to avoid references to classes that do not yet exist.

        Parameters
        ----------
        schema_classes_list : list
            A list of classes as they are found in a schema JSON description.
        """

        for weaviate_class in schema_classes_list:
            self._create_class_with_primitives(weaviate_class)

    def add_class_tenants(self, class_name: str, tenants: List[Tenant]) -> None:
        """
        Add class's tenants in Weaviate.

        Parameters
        ----------
        class_name : str
            The class for which we add tenants.
        tenants : List[Tenant]
            List of Tenants.

        Examples
        --------
        >>> tenants = [ Tenant(name="Tenant1"), Tenant(name="Tenant2") ]
        >>> client.schema.add_class_tenants("class_name", tenants)

        Raises
        ------
        TypeError
            If 'tenants' has not the correct type.
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a non-OK status.
        """

        loaded_tenants = [tenant._to_weaviate_object() for tenant in tenants]

        path = f"/schema/{_capitalize_first_letter(class_name)}/tenants"
        try:
            response = self._connection.post(path=path, weaviate_object=loaded_tenants)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                "Classes tenants may not have been added properly."
            ) from conn_err
        if response.status_code != 200:
            raise UnexpectedStatusCodeException("Add classes tenants", response)

    def remove_class_tenants(self, class_name: str, tenants: List[str]) -> None:
        """
        Remove class's tenants in Weaviate.

        Parameters
        ----------
        class_name : str
            The class for which we remove tenants.
        tenants : List[str]
            List of tenant names to remove from the given class.

        Examples
        --------
        >>> client.schema.remove_class_tenants("class_name", ["Tenant1", "Tenant2"])

        Raises
        ------
        TypeError
            If 'tenants' has not the correct type.
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a non-OK status.
        """
        path = f"/schema/{_capitalize_first_letter(class_name)}/tenants"
        try:
            response = self._connection.delete(path=path, weaviate_object=tenants)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                "Classes tenants may not have been deleted."
            ) from conn_err
        if response.status_code != 200:
            raise UnexpectedStatusCodeException("Delete classes tenants", response)

    def get_class_tenants(self, class_name: str) -> List[Tenant]:
        """Get class's tenants in Weaviate.

        Parameters
        ----------
        class_name : str
            The class for which we get tenants.

        Examples
        --------
        >>> client.schema.get_class_tenants("class_name")

        Raises
        ------
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a non-OK status.
        """
        path = f"/schema/{_capitalize_first_letter(class_name)}/tenants"
        try:
            response = self._connection.get(path=path)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Could not get class tenants.") from conn_err

        tenant_resp = _decode_json_response_list(response, "Get class tenants")
        assert tenant_resp is not None
        return [Tenant._from_weaviate_object(tenant) for tenant in tenant_resp]

    def update_class_tenants(self, class_name: str, tenants: List[Tenant]) -> None:
        """Update class tenants.

        Use this when you want to move tenants from one activity state to another.

        Parameters
        ----------
        class_name : str
            The class for which we update tenants.
        tenants : List[Tenant]
            List of Tenants.

        Examples
        --------
        >>> client.schema.add_class_tenants(
                "class_name",
                [
                    Tenant(activity_status=TenantActivityStatus.HOT, name="Tenant1")),
                    Tenant(activity_status=TenantActivityStatus.COLD, name="Tenant2"))
                    Tenant(name="Tenant3")
                ]
            )
        >>> client.schema.update_class_tenants(
                "class_name",
                [
                    Tenant(activity_status=TenantActivityStatus.COLD, name="Tenant1")),
                    Tenant(activity_status=TenantActivityStatus.HOT, name="Tenant2"))
                ]
            )
        >>> client.schema.get_class_tenants("class_name")
        [
            Tenant(activity_status=TenantActivityStatus.COLD, name="Tenant1")),
            Tenant(activity_status=TenantActivityStatus.HOT, name="Tenant2")),
            Tenant(activity_status=TenantActivityStatus.HOT, name="Tenant3"))
        ]


        Raises
        ------
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a non-OK status.
        """
        path = f"/schema/{_capitalize_first_letter(class_name)}/tenants"
        loaded_tenants = [tenant._to_weaviate_object() for tenant in tenants]
        try:
            response = self._connection.put(path=path, weaviate_object=loaded_tenants)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Could not update class tenants.") from conn_err
        if response.status_code != 200:
            raise UnexpectedStatusCodeException("Update classes tenants", response)


def _property_is_primitive(data_type_list: list) -> bool:
    """
    Check if the property is primitive.

    Parameters
    ----------
    data_type_list : list
        Data types to be checked if are primitive.

    Returns
    -------
    bool
        True if it only consists of primitive data types,
        False otherwise.
    """

    if len(set(data_type_list) - _PRIMITIVE_WEAVIATE_TYPES_SET) == 0:
        return True
    return False


def _get_primitive_properties(properties_list: list) -> list:
    """
    Filter the list of properties for only primitive properties.

    Parameters
    ----------
    properties_list : list
        A list of properties to extract the primitive properties.

    Returns
    -------
    list
        A list of properties containing only primitives.
    """

    primitive_properties = []
    for property_ in properties_list:
        if not _property_is_primitive(property_["dataType"]):
            # property is complex and therefore will be ignored
            continue
        primitive_properties.append(property_)
    return primitive_properties


def _update_nested_dict(dict_1: dict, dict_2: dict) -> dict:
    """
    Update `dict_1` with elements from `dict_2` in a nested manner.
    If a value of a key is a dict, it is going to be updated and not replaced by the whole dict.

    Parameters
    ----------
    dict_1 : dict
        The dictionary to be updated.
    dict_2 : dict
        The dictionary that contains values to be updated.

    Returns
    -------
    dict
        The updated `dict_1`.
    """
    for key, value in dict_2.items():
        if key not in dict_1:
            dict_1[key] = value
            continue
        if isinstance(value, dict):
            _update_nested_dict(dict_1[key], value)
        else:
            dict_1.update({key: value})
    return dict_1



import unittest
from unittest.mock import patch, Mock

from requests.exceptions import ConnectionError as RequestsConnectionError

from test.util import mock_connection_func, check_error_message, check_startswith_error_message
from weaviate.data import DataObject
from weaviate.data.replication import ConsistencyLevel
from weaviate.exceptions import (
    UnexpectedStatusCodeException,
    ObjectAlreadyExistsException,
)


class TestDataObject(unittest.TestCase):
    @patch("weaviate.data.crud_data._get_dict_from_object", side_effect=lambda x: x)
    @patch("weaviate.data.crud_data.get_valid_uuid", side_effect=lambda x: x)
    @patch("weaviate.data.crud_data.get_vector", side_effect=lambda x: x)
    def test_create(self, mock_get_vector, mock_get_valid_uuid, mock_get_dict_from_object):
        """
        Test the `create` method.
        """

        def reset():
            """
            Reset patched objects
            """

            mock_get_valid_uuid.reset_mock()  # reset called
            mock_get_vector.reset_mock()  # reset called
            mock_get_dict_from_object.reset_mock()  # reset_called

        data_object = DataObject(Mock())

        # invalid calls
        class_name_error_message = lambda dt: f"Expected class_name of type str but was: {dt}"
        requests_error_message = "Object was not added to Weaviate."

        # tests
        with self.assertRaises(TypeError) as error:
            data_object.create({"name": "Optimus Prime"}, ["Transformer"])
        check_error_message(self, error, class_name_error_message(list))
        mock_get_dict_from_object.assert_not_called()
        mock_get_vector.assert_not_called()
        mock_get_valid_uuid.assert_not_called()

        # test invalid consistency level
        reset()
        with self.assertRaises(ValueError) as error:
            data_object.create(
                {"name": "Optimus Prime"}, "class", "123", None, consistency_level="TWO"
            )
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()
        mock_get_valid_uuid.assert_called()

        reset()
        mock_obj = mock_connection_func("post", side_effect=RequestsConnectionError("Test!"))
        data_object = DataObject(mock_obj)
        with self.assertRaises(RequestsConnectionError) as error:
            data_object.create({"name": "Alan Greenspan"}, "CoolestPersonEver")
        check_error_message(self, error, requests_error_message)
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()
        mock_get_valid_uuid.assert_not_called()

        reset()
        mock_obj = mock_connection_func("post", status_code=204, return_json={})
        data_object = DataObject(mock_obj)
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            data_object.create({"name": "Alan Greenspan"}, "CoolestPersonEver")
        check_startswith_error_message(self, error, "Creating object")
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()
        mock_get_valid_uuid.assert_not_called()

        reset()
        mock_obj = mock_connection_func(
            "post", status_code=204, return_json={"error": [{"message": "already exists"}]}
        )
        data_object = DataObject(mock_obj)
        with self.assertRaises(ObjectAlreadyExistsException) as error:
            data_object.create({"name": "Alan Greenspan"}, "CoolestPersonEver")
        check_error_message(self, error, "None")
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()
        mock_get_valid_uuid.assert_not_called()

        reset()
        mock_obj = mock_connection_func("post", status_code=204, return_json={})
        data_object = DataObject(mock_obj)
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            data_object.create({"name": "Alan Greenspan"}, "CoolestPersonEver")
        check_startswith_error_message(self, error, "Creating object")
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()
        mock_get_valid_uuid.assert_not_called()

        # # test valid calls
        ## without vector argument
        connection_mock = mock_connection_func("post", return_json={"id": 0}, status_code=200)
        data_object = DataObject(connection_mock)

        object_ = {
            "lyrics": "da da dadadada dada, da da dadadada da, da da dadadada da, da da dadadada da Tequila"
        }
        class_name = "KaraokeSongs"
        vector = [1.0, 2.0]
        id_ = "ae6d51d6-b4ea-5a03-a808-6aae990bdebf"

        rest_object = {"class": class_name, "properties": object_, "id": id_}

        reset()
        uuid = data_object.create(object_, class_name, id_)
        self.assertEqual(uuid, "0")
        connection_mock.post.assert_called_with(
            path="/objects", weaviate_object=rest_object, params={}
        )
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()
        mock_get_valid_uuid.assert_called()

        ## with vector argument
        connection_mock = mock_connection_func("post", return_json={"id": 0}, status_code=200)
        data_object = DataObject(connection_mock)

        object_ = {
            "lyrics": "da da dadadada dada, da da dadadada da, da da dadadada da, da da dadadada da Tequila"
        }
        class_name = "KaraokeSongs"
        vector = [1.0, 2.0]
        id_ = "ae6d51d6-b4ea-5a03-a808-6aae990bdebf"

        rest_object = {"class": class_name, "properties": object_, "vector": vector, "id": id_}

        reset()
        uuid = data_object.create(object_, class_name, id_, vector, "ALL")
        self.assertEqual(uuid, "0")
        connection_mock.post.assert_called_with(
            path="/objects", weaviate_object=rest_object, params={"consistency_level": "ALL"}
        )
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_called()
        mock_get_valid_uuid.assert_called()

        reset()
        # uncapitalized class_names should be capitalized
        uuid = data_object.create(object_, "karaokeSongs", id_, vector)
        self.assertEqual(uuid, "0")
        connection_mock.post.assert_called_with(
            path="/objects", weaviate_object=rest_object, params={}
        )
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_called()
        mock_get_valid_uuid.assert_called()

    @patch("weaviate.data.crud_data._get_dict_from_object", side_effect=lambda x: x)
    @patch("weaviate.data.crud_data.get_vector", side_effect=lambda x: x)
    def test_update(self, mock_get_vector, mock_get_dict_from_object):
        """
        Test the `update` method.
        """
        uuid = "ae6d51d6-b4ea-5a03-a808-6aae990bdebf"
        data_object = DataObject(Mock())

        # error messages
        class_type_error_message = "Class must be type str"
        uuid_type_error_message = (
            lambda dt: f"'uuid' must be of type str or uuid.UUID, but was: {dt}"
        )
        uuid_value_error_message = "Not valid 'uuid' or 'uuid' can not be extracted from value"
        requests_error_message = "Object was not updated."
        unexpected_error_message = "Update of the object not successful"

        with self.assertRaises(ValueError) as error:
            data_object.update({"A": "B"}, "class", uuid, consistency_level="Unknown")
        check_error_message(self, error, "'Unknown' is not a valid ConsistencyLevel")
        mock_get_dict_from_object.assert_not_called()
        mock_get_vector.assert_not_called()

        with self.assertRaises(TypeError) as error:
            data_object.update({"A": "B"}, 35, uuid)
        check_error_message(self, error, class_type_error_message)
        mock_get_dict_from_object.assert_not_called()
        mock_get_vector.assert_not_called()

        with self.assertRaises(TypeError) as error:
            data_object.update({"A": "B"}, "Class", 1238234)
        check_error_message(self, error, uuid_type_error_message(int))
        mock_get_dict_from_object.assert_not_called()
        mock_get_vector.assert_not_called()

        with self.assertRaises(ValueError) as error:
            data_object.update({"A": "B"}, "Class", "NOT-A-valid-uuid")
        check_error_message(self, error, uuid_value_error_message)
        mock_get_dict_from_object.assert_not_called()
        mock_get_vector.assert_not_called()

        mock_obj = mock_connection_func("patch", side_effect=RequestsConnectionError("Test!"))
        data_object = DataObject(mock_obj)
        with self.assertRaises(RequestsConnectionError) as error:
            data_object.update(
                {"name": "Alan Greenspan"},
                "CoolestPersonEver",
                uuid,
            )
        check_error_message(self, error, requests_error_message)
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()

        with self.assertRaises(UnexpectedStatusCodeException) as error:
            mock_obj = mock_connection_func("patch", status_code=200, return_json={})
            data_object = DataObject(mock_obj)
            data_object.update(
                {"name": "Alan Greenspan"},
                "CoolestPersonEver",
                uuid,
            )
        check_startswith_error_message(self, error, unexpected_error_message)
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()

        # test valid calls
        ## without vector argument
        connection_mock = mock_connection_func("patch", status_code=204)
        data_object = DataObject(connection_mock)
        data_object.update({"A": "B"}, "Class", uuid)
        weaviate_obj = {
            "id": uuid,
            "class": "Class",
            "properties": {"A": "B"},
        }
        connection_mock.patch.assert_called_with(
            path="/objects/ae6d51d6-b4ea-5a03-a808-6aae990bdebf",
            weaviate_object=weaviate_obj,
            params={},
        )
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()

        ### with uncapitalized class_name
        connection_mock = mock_connection_func("patch", status_code=204)
        data_object = DataObject(connection_mock)
        data_object.update({"A": "B"}, "class", uuid)
        weaviate_obj = {
            "id": "ae6d51d6-b4ea-5a03-a808-6aae990bdebf",
            "class": "Class",
            "properties": {"A": "B"},
        }
        connection_mock.patch.assert_called_with(
            path="/objects/ae6d51d6-b4ea-5a03-a808-6aae990bdebf",
            weaviate_object=weaviate_obj,
            params={},
        )
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()

        ## with vector argument
        connection_mock = mock_connection_func("patch", status_code=204)
        data_object = DataObject(connection_mock)
        data_object.update(
            {"A": "B"},
            "Class",
            uuid,
            vector=[2.0, 4.0],
            consistency_level=ConsistencyLevel.ONE,
        )
        weaviate_obj = {
            "id": uuid,
            "class": "Class",
            "properties": {"A": "B"},
            "vector": [2.0, 4.0],
        }
        connection_mock.patch.assert_called_with(
            path="/objects/ae6d51d6-b4ea-5a03-a808-6aae990bdebf",
            weaviate_object=weaviate_obj,
            params={"consistency_level": "ONE"},
        )
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_called()

    @patch("weaviate.data.crud_data._get_dict_from_object", side_effect=lambda x: x)
    @patch("weaviate.data.crud_data.get_vector", side_effect=lambda x: x)
    def test_replace(self, mock_get_vector, mock_get_dict_from_object):
        """
        Test the `replace` method.
        """
        uuid = "27be9d8d-1da1-4d52-821f-bc7e2a25247d"
        # test invalid consistency level
        data_object = DataObject(Mock)
        with self.assertRaises(ValueError) as error:
            data_object.replace(
                {"name": "Alan Greenspan"},
                "CoolestPersonEver",
                "27be9d8d-1da1-4d52-821f-bc7e2a25247d",
                consistency_level="Unknown",
            )
        mock_get_dict_from_object.assert_not_called()
        mock_get_vector.assert_not_called()

        # error messages
        requests_error_message = "Object was not replaced."
        unexpected_error_message = "Replace object"

        # test exceptions
        mock_obj = mock_connection_func("put", side_effect=RequestsConnectionError("Test!"))
        data_object = DataObject(mock_obj)
        with self.assertRaises(RequestsConnectionError) as error:
            data_object.replace(
                {"name": "Alan Greenspan"},
                "CoolestPersonEver",
                uuid,
            )
        check_error_message(self, error, requests_error_message)
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()

        mock_obj = mock_connection_func("put", status_code=204, return_json={})
        data_object = DataObject(mock_obj)
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            data_object.replace(
                {"name": "Alan Greenspan"},
                "CoolestPersonEver",
                uuid,
            )
        check_startswith_error_message(self, error, unexpected_error_message)
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()

        # test valid calls
        ## without vector argument
        connection_mock = mock_connection_func("put")
        data_object = DataObject(connection_mock)
        data_object.replace({"A": 2}, "Hero", uuid)
        weaviate_obj = {
            "id": "27be9d8d-1da1-4d52-821f-bc7e2a25247d",
            "class": "Hero",
            "properties": {"A": 2},
        }
        connection_mock.put.assert_called_with(
            path="/objects/27be9d8d-1da1-4d52-821f-bc7e2a25247d",
            weaviate_object=weaviate_obj,
            params={},
        )
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()

        ### with uncapitalized class_name
        connection_mock = mock_connection_func("put")
        data_object = DataObject(connection_mock)
        data_object.replace({"A": 2}, "hero", uuid)
        weaviate_obj = {
            "id": "27be9d8d-1da1-4d52-821f-bc7e2a25247d",
            "class": "Hero",
            "properties": {"A": 2},
        }
        connection_mock.put.assert_called_with(
            path="/objects/27be9d8d-1da1-4d52-821f-bc7e2a25247d",
            weaviate_object=weaviate_obj,
            params={},
        )
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()

        # with vector argument
        connection_mock = mock_connection_func("put")
        data_object = DataObject(connection_mock)
        data_object.replace(
            {"A": 2},
            "Hero",
            uuid,
            vector=[3.0, 5, 7],
            consistency_level="ONE",
        )
        weaviate_obj = {
            "id": "27be9d8d-1da1-4d52-821f-bc7e2a25247d",
            "class": "Hero",
            "properties": {"A": 2},
            "vector": [3.0, 5, 7],
        }
        connection_mock.put.assert_called_with(
            path="/objects/27be9d8d-1da1-4d52-821f-bc7e2a25247d",
            weaviate_object=weaviate_obj,
            params={"consistency_level": "ONE"},
        )
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_called()

    def test_delete(self):
        """
        Test the `delete` method.
        """
        uuid = "b36268d4-a6b5-5274-985f-45f13ce0c642"
        connection_mock = Mock()
        connection_mock.server_version = "1.13.2"
        data_object = DataObject(connection_mock)

        # error messages
        uuid_type_error_message = (
            lambda dt: f"'uuid' must be of type str or uuid.UUID, but was: {dt}"
        )
        uuid_value_error_message = "Not valid 'uuid' or 'uuid' can not be extracted from value"
        requests_error_message = "Object could not be deleted."
        unexpected_error_message = "Delete object"

        with self.assertRaises(TypeError) as error:
            data_object.delete(4)
        check_error_message(self, error, uuid_type_error_message(int))

        with self.assertRaises(ValueError) as error:
            data_object.delete("Hallo World")
        check_error_message(self, error, uuid_value_error_message)

        ## test invalid consistency level
        with self.assertRaises(ValueError) as error:
            data_object.delete(uuid=uuid, consistency_level="Unknown")

        connection_mock = mock_connection_func(
            "delete", side_effect=RequestsConnectionError("Test!")
        )
        data_object = DataObject(connection_mock)
        with self.assertRaises(RequestsConnectionError) as error:
            data_object.delete("b36268d4-a6b5-5274-985f-45f13ce0c642")
        check_error_message(self, error, requests_error_message)

        connection_mock = mock_connection_func("delete", status_code=405)
        data_object = DataObject(connection_mock)
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            data_object.delete("b36268d4-a6b5-5274-985f-45f13ce0c642")
        check_startswith_error_message(self, error, unexpected_error_message)

        # 1. Successfully delete something
        connection_mock = mock_connection_func("delete", status_code=204)
        data_object = DataObject(connection_mock)

        object_id = "b36268d4-a6b5-5274-985f-45f13ce0c642"
        data_object.delete(object_id)
        connection_mock.delete.assert_called_with(
            path="/objects/" + object_id,
            params={},
        )

        data_object.delete(object_id, None, ConsistencyLevel.ALL)
        connection_mock.delete.assert_called_with(
            path="/objects/" + object_id,
            params={"consistency_level": "ALL"},
        )

    def test_get_by_id(self):
        """
        Test the `get_by_id` method.
        """

        data_object = DataObject(Mock())

        mock_get = Mock(return_value="Test")
        data_object.get = mock_get
        data_object.get_by_id(
            uuid="UUID", additional_properties=["Test", "Array"], with_vector=True
        )
        mock_get.assert_called_with(
            uuid="UUID",
            class_name=None,
            additional_properties=["Test", "Array"],
            with_vector=True,
            node_name=None,
            consistency_level=None,
            tenant=None,
        )

        data_object.get_by_id(uuid="UUID2", additional_properties=["Test"], with_vector=False)
        mock_get.assert_called_with(
            uuid="UUID2",
            class_name=None,
            additional_properties=["Test"],
            with_vector=False,
            node_name=None,
            consistency_level=None,
            tenant=None,
        )

        data_object.get_by_id(
            uuid="UUID3",
            additional_properties=["Test"],
            with_vector=False,
            consistency_level=ConsistencyLevel.QUORUM,
        )
        mock_get.assert_called_with(
            uuid="UUID3",
            class_name=None,
            additional_properties=["Test"],
            with_vector=False,
            node_name=None,
            consistency_level=ConsistencyLevel.QUORUM,
            tenant=None,
        )

        data_object.get_by_id(
            uuid="UUID4", additional_properties=["Test"], with_vector=False, node_name="node1"
        )
        mock_get.assert_called_with(
            uuid="UUID4",
            class_name=None,
            additional_properties=["Test"],
            with_vector=False,
            node_name="node1",
            consistency_level=None,
            tenant=None,
        )

        connection_mock = Mock()
        connection_mock.server_version = "1.17.0"
        data_object = DataObject(connection_mock)
        with self.assertRaises(ValueError) as error:
            data_object.get_by_id(
                uuid="UUID4",
                class_name="SomeClass",
                additional_properties=["Test"],
                with_vector=False,
                consistency_level=12345,
            )
            assert "123" in error
        with self.assertRaises(ValueError) as error:
            data_object.get_by_id(
                uuid="UUID4",
                class_name="SomeClass",
                additional_properties=["Test"],
                with_vector=False,
                consistency_level="all",
            )
            assert "all" in error
        with self.assertRaises(ValueError) as error:
            data_object.get_by_id(
                uuid="UUID4",
                class_name="SomeClass",
                additional_properties=["Test"],
                with_vector=False,
                consistency_level={"consistency_level": "ALL"},
            )
            assert "consistency_level" in error

    @patch("weaviate.data.crud_data._get_params")
    def test_get(self, mock_get_params):
        """
        Test the `get` method.
        """

        # error messages
        requests_error_message = "Could not get object/s."
        unexpected_error_message = "Get object/s"

        # test exceptions

        data_object = DataObject(
            mock_connection_func("get", side_effect=RequestsConnectionError("Test!"))
        )
        with self.assertRaises(RequestsConnectionError) as error:
            data_object.get()
        check_error_message(self, error, requests_error_message)

        data_object = DataObject(mock_connection_func("get", status_code=405))
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            data_object.get()
        check_startswith_error_message(self, error, unexpected_error_message)

        # test valid calls
        return_value_get = {"my_key": 12341}
        mock_get_params.return_value = {"include": "test1,test2"}
        connection_mock = mock_connection_func("get", return_json=return_value_get, status_code=200)
        data_object = DataObject(connection_mock)
        result = data_object.get()
        self.assertEqual(result, return_value_get)
        connection_mock.get.assert_called_with(path="/objects", params={"include": "test1,test2"})

        return_value_get = {"my_key": "12341"}
        mock_get_params.return_value = {"include": "test1,test2"}
        connection_mock = mock_connection_func("get", return_json=return_value_get, status_code=200)
        data_object = DataObject(connection_mock)
        result = data_object.get(uuid="1d420c9c98cb11ec9db61e008a366d49")
        self.assertEqual(result, return_value_get)
        connection_mock.get.assert_called_with(
            path="/objects/1d420c9c-98cb-11ec-9db6-1e008a366d49", params={"include": "test1,test2"}
        )

    def test_exists(self):
        """
        Test the `exists` method.
        """
        uuid = "1d420c9c-98cb-11ec-9db6-1e008a366d49"
        # error messages
        requests_error_message = "Could not check if object exist."
        unexpected_error_message = "Object exists"

        # test exceptions
        data_object = DataObject(mock_connection_func("head"))
        with self.assertRaises(ValueError) as error:
            data_object.exists(uuid=uuid, consistency_level="Unknown")

        data_object = DataObject(
            mock_connection_func("head", side_effect=RequestsConnectionError("Test!"))
        )
        with self.assertRaises(RequestsConnectionError) as error:
            data_object.exists(uuid="1d420c9c98cb11ec9db61e008a366d49")
        check_error_message(self, error, requests_error_message)

        data_object = DataObject(mock_connection_func("head", status_code=200))
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            data_object.exists(uuid="1d420c9c98cb11ec9db61e008a366d49")
        check_startswith_error_message(self, error, unexpected_error_message)

        # test valid calls
        connection_mock = mock_connection_func("head", status_code=204)
        data_object = DataObject(connection_mock)
        result = data_object.exists(uuid="1d420c9c98cb11ec9db61e008a366d49")
        self.assertEqual(result, True)
        connection_mock.head.assert_called_with(
            path="/objects/1d420c9c-98cb-11ec-9db6-1e008a366d49",
            params={},
        )

        connection_mock = mock_connection_func("head", status_code=404)
        data_object = DataObject(connection_mock)
        result = data_object.exists(uuid, None, ConsistencyLevel.QUORUM)
        self.assertEqual(result, False)
        connection_mock.head.assert_called_with(
            path="/objects/1d420c9c-98cb-11ec-9db6-1e008a366d49",
            params={"consistency_level": "QUORUM"},
        )

    @patch("weaviate.data.crud_data._get_dict_from_object", side_effect=lambda x: x)
    @patch("weaviate.data.crud_data.get_vector", side_effect=lambda x: x)
    def test_validate(self, mock_get_vector, mock_get_dict_from_object):
        """
        Test the `validate` method.
        """

        data_object = DataObject(Mock())

        # error messages
        uuid_type_error_message = (
            lambda dt: f"'uuid' must be of type str or uuid.UUID, but was: {dt}"
        )
        class_name_error_message = lambda dt: f"Expected class_name of type `str` but was: {dt}"
        requests_error_message = "Object was not validated against Weaviate."
        unexpected_error_message = "Validate object"

        # test exceptions
        with self.assertRaises(TypeError) as error:
            data_object.validate({}, "Name", 1)
        check_error_message(self, error, uuid_type_error_message(int))
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()

        with self.assertRaises(TypeError) as error:
            data_object.validate({}, ["Name"], "73802305-c0da-427e-b21c-d6779a22f35f")
        check_error_message(self, error, class_name_error_message(list))
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()

        data_object = DataObject(
            mock_connection_func("post", side_effect=RequestsConnectionError("Test!"))
        )
        with self.assertRaises(RequestsConnectionError) as error:
            data_object.validate(
                {"name": "Alan Greenspan"},
                "CoolestPersonEver",
                "73802305-c0da-427e-b21c-d6779a22f35f",
            )
        check_error_message(self, error, requests_error_message)
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()

        data_object = DataObject(mock_connection_func("post", status_code=204, return_json={}))
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            data_object.validate(
                {"name": "Alan Greenspan"},
                "CoolestPersonEver",
                "73802305-c0da-427e-b21c-d6779a22f35f",
            )
        check_startswith_error_message(self, error, unexpected_error_message)
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()

        # test valid calls
        # test for status_code 200 without vector argument
        connection_mock = mock_connection_func("post", status_code=200)
        data_object = DataObject(connection_mock)

        response = data_object.validate({"A": 2}, "Hero", "27be9d8d-1da1-4d52-821f-bc7e2a25247d")
        self.assertEqual(response, {"error": None, "valid": True})

        weaviate_obj = {
            "id": "27be9d8d-1da1-4d52-821f-bc7e2a25247d",
            "class": "Hero",
            "properties": {"A": 2},
        }
        connection_mock.post.assert_called_with(
            path="/objects/validate", weaviate_object=weaviate_obj
        )
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()

        ### with uncapitalized class_name
        connection_mock = mock_connection_func("post", status_code=200)
        data_object = DataObject(connection_mock)

        response = data_object.validate({"A": 2}, "hero", "27be9d8d-1da1-4d52-821f-bc7e2a25247d")
        self.assertEqual(response, {"error": None, "valid": True})

        weaviate_obj = {
            "id": "27be9d8d-1da1-4d52-821f-bc7e2a25247d",
            "class": "Hero",
            "properties": {"A": 2},
        }
        connection_mock.post.assert_called_with(
            path="/objects/validate", weaviate_object=weaviate_obj
        )
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()

        # test for status_code 422
        connection_mock = mock_connection_func(
            "post", status_code=422, return_json={"error": "Not OK!"}
        )
        data_object = DataObject(connection_mock)

        response = data_object.validate({"A": 2}, "Hero", "27be9d8d-1da1-4d52-821f-bc7e2a25247d")
        self.assertEqual(response, {"error": "Not OK!", "valid": False})

        weaviate_obj = {
            "id": "27be9d8d-1da1-4d52-821f-bc7e2a25247d",
            "class": "Hero",
            "properties": {"A": 2},
        }
        connection_mock.post.assert_called_with(
            path="/objects/validate", weaviate_object=weaviate_obj
        )
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_not_called()

        # test for status_code 200 with vector argument
        connection_mock = mock_connection_func("post", status_code=200)
        data_object = DataObject(connection_mock)

        response = data_object.validate(
            {"A": 2}, "Hero", "27be9d8d-1da1-4d52-821f-bc7e2a25247d", vector=[-9.8, 6.66]
        )
        self.assertEqual(response, {"error": None, "valid": True})

        weaviate_obj = {
            "id": "27be9d8d-1da1-4d52-821f-bc7e2a25247d",
            "class": "Hero",
            "properties": {"A": 2},
            "vector": [-9.8, 6.66],
        }
        connection_mock.post.assert_called_with(
            path="/objects/validate", weaviate_object=weaviate_obj
        )
        mock_get_dict_from_object.assert_called()
        mock_get_vector.assert_called()

    def test__get_params(self):
        """
        Test the `_get_params` function.
        """

        from weaviate.data.crud_data import _get_params

        # error messages
        type_error_message = lambda dt: f"Additional properties must be of type list but are {dt}"

        with self.assertRaises(TypeError) as error:
            _get_params("Test", False)
        check_error_message(self, error, type_error_message(str))

        self.assertEqual(_get_params(["test1", "test2"], False), {"include": "test1,test2"})
        self.assertEqual(_get_params(None, True), {"include": "vector"})
        self.assertEqual(_get_params([], True), {"include": "vector"})
        self.assertEqual(_get_params(["test1", "test2"], True), {"include": "test1,test2,vector"})



"""
Contextionary class definition.
"""
from requests.exceptions import ConnectionError as RequestsConnectionError

from weaviate.connect import Connection
from weaviate.exceptions import UnexpectedStatusCodeException
from weaviate.util import _decode_json_response_dict


class Contextionary:
    """
    Contextionary class used to add extend the Weaviate contextionary module
    or to get vector/s of a specific concept.
    """

    def __init__(self, connection: Connection):
        """
        Initialize a Contextionary class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running Weaviate instance.
        """

        self._connection = connection

    def extend(self, concept: str, definition: str, weight: float = 1.0) -> None:
        """
        Extend the text2vec-contextionary with new concepts

        Parameters
        ----------
        concept : str
            The new concept that should be added that is not in the Weaviate
            or needs to be updated, e.g. an abbreviation.
        definition : str
            The definition of the new concept.
        weight : float, optional
            The weight of the new definition compared to the old one,
            must be in-between the interval [0.0; 1.0], by default 1.0

        Examples
        --------
        >>> client.contextionary.extend(
        ...     concept = 'palantir',
        ...     definition = 'spherical stone objects used for communication in Middle-earth'
        ... )


        Raises
        ------
        TypeError
            If an argument is not of an appropriate type.
        ValueError
            If 'weight' is outside the interval [0.0; 1.0].
        requests.ConnectionError
            If text2vec-contextionary could not be extended.
        weaviate.UnexpectedStatusCodeException
            If the network connection to weaviate fails.
        """

        if not isinstance(concept, str):
            raise TypeError("Concept must be string")
        if not isinstance(definition, str):
            raise TypeError("Definition must be string")
        if not isinstance(weight, float):
            raise TypeError("Weight must be float")

        if weight > 1.0 or weight < 0.0:
            raise ValueError("Weight out of limits 0.0 <= weight <= 1.0")

        extension = {"concept": concept, "definition": definition, "weight": weight}

        try:
            response = self._connection.post(
                path="/modules/text2vec-contextionary/extensions",
                weaviate_object=extension,
            )
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                "text2vec-contextionary could not be extended."
            ) from conn_err
        if response.status_code == 200:
            # Successfully extended
            return
        raise UnexpectedStatusCodeException("Extend text2vec-contextionary", response)

    def get_concept_vector(self, concept: str) -> dict:
        """
        Retrieves the vector representation of the given concept.

        Parameters
        ----------
        concept : str
            Concept for which the vector should be retrieved.
            May be camelCase for word combinations.

        Examples
        --------
        >>> client.contextionary.get_concept_vector('king')
        {
            "individualWords": [
                {
                "info": {
                    "nearestNeighbors": [
                    {
                        "word": "king"
                    },
                    {
                        "distance": 5.7498446,
                        "word": "kings"
                    },
                    ...,
                    {
                        "distance": 6.1396513,
                        "word": "queen"
                    }
                    ],
                    "vector": [
                    -0.68988,
                    ...,
                    -0.561865
                    ]
                },
                "present": true,
                "word": "king"
                }
            ]
        }

        Returns
        -------
        dict
            A dictionary containing info and the vector/s of the concept.
            The vector might be empty if the text2vec-contextionary does not contain it.

        Raises
        ------
        requests.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        path = "/modules/text2vec-contextionary/concepts/" + concept
        try:
            response = self._connection.get(path=path)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                "text2vec-contextionary vector was not retrieved."
            ) from conn_err
        else:
            res = _decode_json_response_dict(response, "text2vec-contextionary vector")
            assert res is not None
            return res



import hashlib
import os
import platform
import re
import socket
import stat
import subprocess
import tarfile
import time
import urllib.request
import warnings
import zipfile
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional

import requests
import validators  # type: ignore

from weaviate import exceptions
from weaviate.exceptions import WeaviateStartUpError
from weaviate.util import _decode_json_response_dict

DEFAULT_BINARY_PATH = str(Path.home() / ".cache/weaviate-embedded/")
DEFAULT_PERSISTENCE_DATA_PATH = str(Path.home() / ".local/share/weaviate")
GITHUB_RELEASE_DOWNLOAD_URL = "https://github.com/weaviate/weaviate/releases/download/"

DEFAULT_PORT = 8079
DEFAULT_GRPC_PORT = 50060


@dataclass
class EmbeddedOptions:
    persistence_data_path: str = os.environ.get("XDG_DATA_HOME", DEFAULT_PERSISTENCE_DATA_PATH)
    binary_path: str = os.environ.get("XDG_CACHE_HOME", DEFAULT_BINARY_PATH)
    version: str = "1.21.4"
    port: int = DEFAULT_PORT
    hostname: str = "127.0.0.1"
    additional_env_vars: Optional[Dict[str, str]] = None
    grpc_port: int = DEFAULT_GRPC_PORT


def get_random_port() -> int:
    sock = socket.socket()
    sock.bind(("", 0))
    port_num = int(sock.getsockname()[1])
    sock.close()
    return port_num


class EmbeddedDB:
    def __init__(self, options: EmbeddedOptions) -> None:
        self.data_bind_port = get_random_port()
        self.options = options
        self.grpc_port: int = options.grpc_port
        self.process: Optional[subprocess.Popen[bytes]] = None
        self.ensure_paths_exist()
        self.check_supported_platform()
        self._parsed_weaviate_version = ""
        # regular expression to detect a version number: v[one digit].[1-2 digits].[1-2 digits]
        # optionally there can be a "-rc/alpha/beta.[1-2 digits]"
        # nothing in front or back
        version_pattern = re.compile(
            r"^\d\.\d{1,2}\.\d{1,2}?(-rc\.\d{1,2}|-beta\.\d{1,2}|-alpha\.\d{1,2}|$)$"
        )

        valid_url = validators.url(self.options.version)
        if isinstance(valid_url, validators.ValidationError):
            valid_url = validators.url(self.options.version, simple_host=True)  # for localhost
        if valid_url:
            if not self.options.version.endswith(".tar.gz") and not self.options.version.endswith(
                ".zip"
            ):
                raise exceptions.WeaviateEmbeddedInvalidVersion(self.options.version)

            # for GitHub urls we can parse the version from the url
            if self.options.version.startswith(GITHUB_RELEASE_DOWNLOAD_URL):
                # replace with str.removeprefix() after 3.8 has been deprecated
                self._parsed_weaviate_version = self.options.version[
                    len(GITHUB_RELEASE_DOWNLOAD_URL) :
                ].split("/")[0]
            self._download_url = self.options.version
        elif version_pattern.match(self.options.version):
            version_tag = "v" + self.options.version
            self._parsed_weaviate_version = version_tag
            self._set_download_url_from_version_tag(version_tag)
        elif self.options.version == "latest":
            response = requests.get(
                "https://api.github.com/repos/weaviate/weaviate/releases/latest"
            )
            latest = _decode_json_response_dict(response, "get tag of latest weaviate release")
            assert latest is not None
            self._set_download_url_from_version_tag(latest["tag_name"])
        else:
            raise exceptions.WeaviateEmbeddedInvalidVersion(self.options.version)

    def _set_download_url_from_version_tag(self, version: str) -> None:
        if platform.system() == "Darwin":
            machine_type = "all"
            package_format = "zip"
        else:
            machine_type = platform.machine()
            if machine_type == "x86_64":
                machine_type = "amd64"
            elif machine_type == "aarch64":
                machine_type = "arm64"
            package_format = "tar.gz"

        self._download_url = (
            GITHUB_RELEASE_DOWNLOAD_URL
            + version
            + "/weaviate-"
            + version
            + "-"
            + platform.system()
            + "-"
            + machine_type
            + "."
            + package_format
        )

    def __del__(self) -> None:
        self.stop()

    def ensure_paths_exist(self) -> None:
        Path(self.options.binary_path).mkdir(parents=True, exist_ok=True)
        Path(self.options.persistence_data_path).mkdir(parents=True, exist_ok=True)

    def ensure_weaviate_binary_exists(self) -> None:
        self._weaviate_binary_path = Path(
            self.options.binary_path,
            "weaviate-"
            + self._parsed_weaviate_version
            + "-"
            + str(hashlib.sha256(self.options.version.encode("utf-8")).hexdigest()),
        )
        if not self._weaviate_binary_path.exists():
            print(
                f"Binary {self.options.binary_path} did not exist. Downloading binary from {self._download_url}"
            )
            if self._download_url.endswith(".tar.gz"):
                tar_filename = Path(self.options.binary_path, "tmp_weaviate.tgz")
                urllib.request.urlretrieve(self._download_url, tar_filename)
                binary_tar = tarfile.open(tar_filename)
                binary_tar.extract("weaviate", path=Path(self.options.binary_path))
                tar_filename.unlink()
            else:
                assert self._download_url.endswith(".zip")
                zip_filename = Path(self.options.binary_path, "tmp_weaviate.zip")
                urllib.request.urlretrieve(self._download_url, zip_filename)
                with zipfile.ZipFile(zip_filename, "r") as zip_ref:
                    zip_ref.extract("weaviate", path=Path(self.options.binary_path))

            (Path(self.options.binary_path) / "weaviate").rename(self._weaviate_binary_path)

            # Ensuring weaviate binary is executable
            self._weaviate_binary_path.chmod(
                self._weaviate_binary_path.stat().st_mode | stat.S_IEXEC
            )

    def is_listening(self) -> bool:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.connect((self.options.hostname, self.options.port))
            s.close()
            return True
        except (socket.error, ConnectionRefusedError):
            s.close()
            return False

    def wait_till_listening(self) -> None:
        seconds = 30
        sleep_interval = 0.1
        retries = int(seconds / sleep_interval)
        while self.is_listening() is False and retries > 0:
            time.sleep(sleep_interval)
            retries -= 1
        if retries == 0:
            raise WeaviateStartUpError(
                f"Embedded DB did not start listening on port {self.options.port} within {seconds} seconds"
            )

    @staticmethod
    def check_supported_platform() -> None:
        if platform.system() in ["Windows"]:
            raise WeaviateStartUpError(
                f"{platform.system()} is not supported with EmbeddedDB. Please upvote this feature request if "
                f"you want this: https://github.com/weaviate/weaviate/issues/3315"
            )

    def start(self) -> None:
        if self.is_listening():
            print(f"embedded weaviate is already listening on port {self.options.port}")
            return

        self.ensure_weaviate_binary_exists()
        my_env = os.environ.copy()

        my_env.setdefault("AUTHENTICATION_ANONYMOUS_ACCESS_ENABLED", "true")
        my_env.setdefault("QUERY_DEFAULTS_LIMIT", "20")
        my_env.setdefault("PERSISTENCE_DATA_PATH", self.options.persistence_data_path)
        # Bug with weaviate requires setting gossip and data bind port
        my_env.setdefault("CLUSTER_GOSSIP_BIND_PORT", str(get_random_port()))
        my_env.setdefault("GRPC_PORT", str(self.grpc_port))

        my_env.setdefault(
            "ENABLE_MODULES",
            "text2vec-openai,text2vec-cohere,text2vec-huggingface,ref2vec-centroid,generative-openai,qna-openai,"
            "reranker-cohere",
        )

        # have a deterministic hostname in case of changes in the network name. This allows to run multiple parallel
        # instances
        my_env.setdefault("CLUSTER_HOSTNAME", f"Embedded_at_{self.options.port}")

        if self.options.additional_env_vars is not None:
            my_env.update(self.options.additional_env_vars)

        # filter warning about running processes.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", ResourceWarning)
            process = subprocess.Popen(
                [
                    f"{self._weaviate_binary_path}",
                    "--host",
                    self.options.hostname,
                    "--port",
                    str(self.options.port),
                    "--scheme",
                    "http",
                ],
                env=my_env,
            )
            self.process = process
        print(f"Started {self.options.binary_path}: process ID {self.process.pid}")
        self.wait_till_listening()

    def stop(self) -> None:
        if self.process is not None:
            try:
                self.process.terminate()
                self.process.wait()
            except ProcessLookupError:
                print(
                    f"""Tried to stop embedded weaviate process {self.process.pid}. Process was not found. So not doing
                    anything"""
                )

    def ensure_running(self) -> None:
        if not self.is_listening():
            print(
                f"Embedded weaviate wasn't listening on port {self.options.port}, so starting embedded weaviate again"
            )
            self.start()



import uuid as uuid_lib
from typing import Any, Dict, List, Literal, Optional, cast, overload

from weaviate.collections.classes.filters import (
    _Filters,
    _FilterAnd,
    _FilterOr,
    _FilterValue,
    FilterValues,
)
from weaviate.util import _datetime_to_string
from weaviate.types import TIME
from weaviate.proto.v1 import search_get_pb2


class _FilterToGRPC:
    @overload
    @staticmethod
    def convert(weav_filter: Literal[None]) -> None:
        ...

    @overload
    @staticmethod
    def convert(weav_filter: _Filters) -> search_get_pb2.Filters:
        ...

    @staticmethod
    def convert(weav_filter: Optional[_Filters]) -> Optional[search_get_pb2.Filters]:
        if weav_filter is None:
            return None
        if isinstance(weav_filter, _FilterValue):
            return _FilterToGRPC.__value_filter(weav_filter)
        else:
            return _FilterToGRPC.__and_or_filter(weav_filter)

    @staticmethod
    def __value_filter(weav_filter: _FilterValue) -> search_get_pb2.Filters:
        return search_get_pb2.Filters(
            operator=weav_filter.operator._to_grpc(),
            value_text=_FilterToGRPC.__filter_to_text(weav_filter.value),
            value_int=weav_filter.value if isinstance(weav_filter.value, int) else None,
            value_boolean=weav_filter.value if isinstance(weav_filter.value, bool) else None,  # type: ignore
            value_number=weav_filter.value if isinstance(weav_filter.value, float) else None,
            value_int_array=_FilterToGRPC.__filter_to_int_list(weav_filter.value),
            value_number_array=_FilterToGRPC.__filter_to_float_list(weav_filter.value),
            value_text_array=_FilterToGRPC.__filter_to_text_list(weav_filter.value),
            value_boolean_array=_FilterToGRPC.__filter_to_bool_list(weav_filter.value),
            on=weav_filter.path if isinstance(weav_filter.path, list) else [weav_filter.path],
        )

    @staticmethod
    def __filter_to_text(value: FilterValues) -> Optional[str]:
        if not (
            isinstance(value, TIME) or isinstance(value, str) or isinstance(value, uuid_lib.UUID)
        ):
            return None

        if isinstance(value, str):
            return value

        if isinstance(value, uuid_lib.UUID):
            return str(value)

        return _datetime_to_string(value)

    @staticmethod
    def __filter_to_text_list(value: FilterValues) -> Optional[search_get_pb2.TextArray]:
        if not isinstance(value, list) or not (
            isinstance(value[0], TIME)
            or isinstance(value[0], str)
            or isinstance(value[0], uuid_lib.UUID)
        ):
            return None

        if isinstance(value[0], str):
            value_list = value
        elif isinstance(value[0], uuid_lib.UUID):
            value_list = [str(uid) for uid in value]
        else:
            dates = cast(List[TIME], value)
            value_list = [_datetime_to_string(date) for date in dates]

        return search_get_pb2.TextArray(values=cast(List[str], value_list))

    @staticmethod
    def __filter_to_bool_list(value: FilterValues) -> Optional[search_get_pb2.BooleanArray]:
        if not isinstance(value, list) or not isinstance(value[0], bool):
            return None

        return search_get_pb2.BooleanArray(values=cast(List[bool], value))

    @staticmethod
    def __filter_to_float_list(value: FilterValues) -> Optional[search_get_pb2.NumberArray]:
        if not isinstance(value, list) or not isinstance(value[0], float):
            return None

        return search_get_pb2.NumberArray(values=cast(List[float], value))

    @staticmethod
    def __filter_to_int_list(value: FilterValues) -> Optional[search_get_pb2.IntArray]:
        if not isinstance(value, list) or not isinstance(value[0], int):
            return None

        return search_get_pb2.IntArray(values=cast(List[int], value))

    @staticmethod
    def __and_or_filter(weav_filter: _Filters) -> Optional[search_get_pb2.Filters]:
        assert isinstance(weav_filter, _FilterAnd) or isinstance(weav_filter, _FilterOr)
        return search_get_pb2.Filters(
            operator=weav_filter.operator._to_grpc(),
            filters=[
                filter_
                for single_filter in weav_filter.filters
                if (filter_ := _FilterToGRPC.convert(single_filter)) is not None
            ],
        )


class _FilterToREST:
    @overload
    @staticmethod
    def convert(weav_filter: Literal[None]) -> None:
        ...

    @overload
    @staticmethod
    def convert(weav_filter: _Filters) -> Dict[str, Any]:
        ...

    @staticmethod
    def convert(weav_filter: Optional[_Filters]) -> Optional[Dict[str, Any]]:
        if weav_filter is None:
            return None
        if isinstance(weav_filter, _FilterValue):
            return _FilterToREST.__value_filter(weav_filter)
        else:
            return _FilterToREST.__and_or_filter(weav_filter)

    @staticmethod
    def __value_filter(weav_filter: _FilterValue) -> Dict[str, Any]:
        return {
            "operator": weav_filter.operator.value,
            "path": weav_filter.path if isinstance(weav_filter.path, list) else [weav_filter.path],
            **_FilterToREST.__parse_filter(weav_filter.value),
        }

    @staticmethod
    def __parse_filter(value: FilterValues) -> Dict[str, Any]:
        if isinstance(value, str):
            return {"valueText": value}
        if isinstance(value, uuid_lib.UUID):
            return {"valueText": str(value)}
        if isinstance(value, TIME):
            return {"valueDate": _datetime_to_string(value)}
        if isinstance(value, bool):
            return {"valueBoolean": value}
        if isinstance(value, int):
            return {"valueInt": value}
        if isinstance(value, float):
            return {"valueNumber": value}
        if isinstance(value, list):
            if isinstance(value[0], str):
                return {"valueTextArray": value}
            if isinstance(value[0], uuid_lib.UUID):
                return {"valueTextArray": [str(val) for val in value]}
            if isinstance(value[0], TIME):
                return {"valueDateArray": [_datetime_to_string(cast(TIME, val)) for val in value]}
            if isinstance(value[0], bool):
                return {"valueBooleanArray": value}
            if isinstance(value[0], int):
                return {"valueIntArray": value}
            if isinstance(value[0], float):
                return {"valueNumberArray": value}
        raise ValueError(f"Unknown filter value type: {type(value)}")

    @staticmethod
    def __and_or_filter(weav_filter: _Filters) -> Dict[str, Any]:
        assert isinstance(weav_filter, _FilterAnd) or isinstance(weav_filter, _FilterOr)
        return {
            "operator": weav_filter.operator.value,
            "operands": [
                filter_
                for single_filter in weav_filter.filters
                if (filter_ := _FilterToREST.convert(single_filter)) is not None
            ],
        }



import uuid

import weaviate
from mock_tests.conftest import MOCK_SERVER_URL


def test_manual_batching_warning_object(recwarn, weaviate_mock):
    weaviate_mock.expect_request("/v1/batch/objects").respond_with_json([])

    client = weaviate.Client(MOCK_SERVER_URL)

    client.batch.configure(batch_size=None, dynamic=False)
    client.batch.add_data_object({}, "ExistingClass")
    client.batch.create_objects()

    assert len(recwarn) == 1
    w = recwarn.pop()
    assert issubclass(w.category, DeprecationWarning)
    assert str(w.message).startswith("Dep002")


def test_manual_batching_warning_ref(recwarn, weaviate_mock):
    weaviate_mock.expect_request("/v1/batch/references").respond_with_json([])

    client = weaviate.Client(MOCK_SERVER_URL)
    client.batch.configure(batch_size=None, dynamic=False)

    client.batch.add_reference(
        str(uuid.uuid4()), "NonExistingClass", "existsWith", str(uuid.uuid4()), "OtherClass"
    )
    client.batch.create_references()

    assert len(recwarn) == 1
    w = recwarn.pop()
    assert issubclass(w.category, DeprecationWarning)
    assert str(w.message).startswith("Dep002")
